import contextlib
import os
import unittest
import gzip
import importlib.resources
from contextlib import nullcontext

from pysylph import Database, Sketcher, SampleSketch


def every_n(it, n=4):
    try:
        while True:
            yield next(it)
            for i in range(n - 1):
                next(it)
    except StopIteration:
        return


def read_fasta_seq(file):
    file.readline()  # skip header
    return "".join(map(str.strip, file))


def read_fastq(file):
    file.readline()  # skip header
    return [line.strip() for line in every_n(file, n=4)]


@contextlib.contextmanager
def get_path(name):
    if hasattr(importlib.resources, "files"):
        handler = nullcontext(importlib.resources.files(__package__).joinpath(name))
    else:
        handler = importlib.resources.path(__package__, name)
    with handler as f:
        yield f


class TestSketcher(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # load reference database
        with get_path("clodf13.syldb") as path:
            cls.db = Database.load(path)
        # load seq
        with get_path("clodf13.fna.gz") as path:
            with gzip.open(path, "rt") as f:
                cls.seq = read_fasta_seq(f)

    def test_sketch_genome_str(self):
        sketcher = Sketcher()
        sketch = sketcher.sketch_genome("CloDF13", [self.seq])
        self.assertEqual(sketch.name, "CloDF13")
        self.assertEqual(sketch.k, self.db[0].k)
        self.assertEqual(sketch.c, self.db[0].c)
        self.assertEqual(sketch.kmers, self.db[0].kmers)

    def test_invalid_kmer(self):
        self.assertRaises(ValueError, Sketcher, k=8)

    def test_sketch_genome_bytes(self):
        b = self.seq.encode()
        sketcher = Sketcher()
        sketch = sketcher.sketch_genome("CloDF13", [b])
        self.assertEqual(sketch.name, "CloDF13")
        self.assertEqual(sketch.k, self.db[0].k)
        self.assertEqual(sketch.c, self.db[0].c)
        self.assertEqual(sketch.kmers, self.db[0].kmers)

    def test_sketch_genome_memoryview(self):
        b = self.seq.encode()
        m = memoryview(b)
        sketcher = Sketcher()
        sketch = sketcher.sketch_genome("CloDF13", [m])
        self.assertEqual(sketch.name, "CloDF13")
        self.assertEqual(sketch.k, self.db[0].k)
        self.assertEqual(sketch.c, self.db[0].c)
        self.assertEqual(sketch.kmers, self.db[0].kmers)

    def test_sketch_single(self):
        with get_path("k12.R1.fq.gz") as path:
            with gzip.open(path, "rt") as f:
                r1 = read_fastq(f)

        sketcher = Sketcher()
        sketch = sketcher.sketch_single("k12", r1)

        with get_path("k12.R1.sylsp") as path:
            expected = SampleSketch.load(path)
        self.assertEqual(sketch.kmer_counts, expected.kmer_counts)

    def test_sketch_paired(self):
        with get_path("k12.R1.fq.gz") as path:
            with gzip.open(path, "rt") as f:
                r1 = read_fastq(f)
        with get_path("k12.R2.fq.gz") as path:
            with gzip.open(path, "rt") as f:
                r2 = read_fastq(f)

        sketcher = Sketcher()
        sketch = sketcher.sketch_paired("k12", r1, r2)

        with get_path("k12.paired.sylsp") as path:
            expected = SampleSketch.load(path)
        self.assertEqual(sketch.kmer_counts, expected.kmer_counts)



import abc
import typing
from os import PathLike
from typing import List, Sequence, Iterable, Iterator, BinaryIO, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

K = Literal[21, 31]
SEQUENCE = Union[str, bytes, bytearray, memoryview]

class Sketch(abc.ABC):
    @property
    @abc.abstractmethod
    def c(self) -> int: ...
    @property
    @abc.abstractmethod
    def k(self) -> K: ...

class GenomeSketch(Sketch):
    @property
    def k(self) -> K: ...
    @property
    def c(self) -> int: ...
    @property
    def name(self) -> str: ...
    @property
    def description(self) -> str: ...
    @property
    def genome_size(self) -> int: ...
    @property
    def min_spacing(self) -> int: ...
    @property
    def kmers(self) -> List[int]: ...

class Database(Sequence[GenomeSketch]):
    def __init__(self, items: Iterable[GenomeSketch] = ()) -> None: ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> GenomeSketch: ...
    @classmethod
    def load(cls, file: Union[str, PathLike, BinaryIO]) -> Database: ...
    def dump(self, path: Union[str, PathLike]) -> None: ...

class DatabaseReader(Iterator[GenomeSketch]):
    def __init__(self, path: Union[str, PathLike]) -> None: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> DatabaseReader: ...
    def __next__(self) -> GenomeSketch: ...

class SampleSketch(Sketch):
    @property
    def k(self) -> K: ...
    @property
    def c(self) -> int: ...
    @classmethod
    def load(cls, file: Union[str, PathLike, BinaryIO]) -> SampleSketch: ...
    def dump(self, path: Union[str, PathLike]) -> None: ...

class AniResult:
    @property
    def genome_sketch(self) -> GenomeSketch: ...
    @property
    def ani(self) -> float: ...
    @property
    def ani_naive(self) -> float: ...
    @property
    def coverage(self) -> float: ...

class ProfileResult(AniResult):
    @property
    def sequence_abundance(self) -> float: ...
    @property
    def taxonomic_abundance(self) -> float: ...
    @property
    def kmers_reassigned(self) -> int: ...

class Sketcher:
    def __init__(
        self, *, c: int = 200, k: K = 31, deduplicate: bool = True, fpr: float = 0.0001
    ) -> None: ...
    def sketch_genome(
        self, name: str, contigs: Iterable[SEQUENCE], profiling: bool = True
    ) -> GenomeSketch: ...
    def sketch_single(self, name: str, reads: Iterable[SEQUENCE]) -> SampleSketch: ...
    def sketch_paired(
        self, name: str, r1: Iterable[SEQUENCE], r2: Iterable[SEQUENCE]
    ) -> SampleSketch: ...

class Profiler:
    def __init__(
        self,
        database: Database,
        *,
        minimum_ani: Optional[float],
        seq_id: Optional[float],
        estimate_unknown: bool = False,
        min_number_kmers: int = 50,
    ) -> None: ...
    def query(self, sample: SampleSketch) -> List[AniResult]: ...
    def profile(self, sample: SampleSketch) -> List[ProfileResult]: ...



import os
import unittest
import gzip
import importlib.resources
from contextlib import nullcontext

from pysylph import Database, SampleSketch, Profiler


class TestProfiler(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # load reference database
        if hasattr(importlib.resources, "files"):
            handler = nullcontext(
                importlib.resources.files(__package__).joinpath("ecoli.syldb")
            )
        else:
            handler = importlib.resources.path(__package__, "ecoli.syldb")
        with handler as path:
            cls.db = Database.load(path)
        # load seq
        if hasattr(importlib.resources, "files"):
            handler = nullcontext(
                importlib.resources.files(__package__).joinpath("o157.sylsp")
            )
        else:
            handler = importlib.resources.path(__package__, "o157.sylsp")
        with handler as path:
            cls.sample = SampleSketch.load(path)

    def test_query(self):
        # sketch sequence
        profiler = Profiler(self.db)
        results = profiler.query(self.sample)

        self.assertEqual(len(results), 3)
        self.assertAlmostEqual(results[0].ani, 99.73, places=2)
        self.assertAlmostEqual(results[0].ani_naive, 96.02, places=2)

        self.assertAlmostEqual(results[1].ani, 98.25, places=2)
        self.assertAlmostEqual(results[1].ani_naive, 94.29, places=2)

        self.assertAlmostEqual(results[2].ani, 98.16, places=2)
        self.assertAlmostEqual(results[2].ani_naive, 94.26, places=2)

    def test_profile(self):
        # sketch sequence
        profiler = Profiler(self.db)
        results = profiler.profile(self.sample)

        self.assertEqual(len(results), 1)
        self.assertAlmostEqual(results[0].ani, 99.73, places=2)
        self.assertAlmostEqual(results[0].ani_naive, 96.02, places=2)
        self.assertAlmostEqual(results[0].sequence_abundance, 100.0, places=2)
        self.assertAlmostEqual(results[0].taxonomic_abundance, 100.0, places=2)
        self.assertAlmostEqual(results[0].kmers_reassigned, 0)



from . import test_database, test_doctest, test_sketcher, test_profiler


def load_tests(loader, suite, pattern):
    suite.addTests(loader.loadTestsFromModule(test_database))
    suite.addTests(loader.loadTestsFromModule(test_doctest))
    suite.addTests(loader.loadTestsFromModule(test_sketcher))
    suite.addTests(loader.loadTestsFromModule(test_profiler))
    return suite



from . import lib
from .lib import (
    Sketcher,
    Profiler,
    Database,
    DatabaseFile,
    Sketch,
    GenomeSketch,
    SampleSketch,
    AniResult,
    ProfileResult,
)

__version__ = lib.__version__
__author__ = lib.__author__
__doc__ = lib.__doc__
# __build__ = lib.__build__
__all__ = [
    "Sketcher",
    "Profiler",
    "Database",
    "DatabaseFile",
    "Sketch",
    "GenomeSketch",
    "SampleSketch",
    "AniResult",
    "ProfileResult",
]



import os
import unittest
import importlib.resources
from contextlib import nullcontext

from pysylph import Database


class TestDatabase(unittest.TestCase):

    def test_init_empty(self):
        db1 = Database()
        self.assertEqual(len(db1), 0)
        db2 = Database([])
        self.assertEqual(len(db2), 0)

    def test_init_type_error(self):
        with self.assertRaises(TypeError):
            db = Database(123)

    def test_load_filename(self):
        if hasattr(importlib.resources, "files"):
            handler = nullcontext(
                importlib.resources.files(__package__).joinpath("ecoli.syldb")
            )
        else:
            handler = importlib.resources.path(__package__, "ecoli.syldb")
        with handler as path:
            database = Database.load(os.fspath(path))
        self.assertEqual(len(database), 3)

    def test_load_path(self):
        if hasattr(importlib.resources, "files"):
            handler = nullcontext(
                importlib.resources.files(__package__).joinpath("ecoli.syldb")
            )
        else:
            handler = importlib.resources.path(__package__, "ecoli.syldb")
        with handler as path:
            database = Database.load(path)
        self.assertEqual(len(database), 3)

    def test_load_file_object(self):
        if hasattr(importlib.resources, "files"):
            handler = (
                importlib.resources.files(__package__)
                .joinpath("ecoli.syldb")
                .open("rb")
            )
        else:
            handler = importlib.resources.open_binary(__package__, "ecoli.syldb")
        with handler as f:
            database = Database.load(f)
        self.assertEqual(len(database), 3)
